package project

import (
	"encoding/json"
	"fmt"
	adminSdk "github.com/friendsofshopware/go-shopware-admin-api-sdk"
	log "github.com/sirupsen/logrus"
	"io/ioutil"
	"os"
	"path/filepath"
	"shopware-cli/shop"
)

type MailTemplateSync struct{}

func (s MailTemplateSync) Push(ctx adminSdk.ApiContext, client *adminSdk.Client, config *shop.Config, operation *ConfigSyncOperation) error {
	mailTemplates, err := fetchAllMailTemplates(ctx, client)
	if err != nil {
		return err
	}

	mailUpdates := make([]map[string]interface{}, 0)

	for _, external := range mailTemplates.Data {
		for _, configEntry := range config.Sync.MailTemplate {
			if external.Id == configEntry.Id {
				mailUpdate := make(map[string]interface{})
				mailUpdate["id"] = configEntry.Id
				translationUpdates := make(map[string]map[string]interface{})

				for _, translation := range external.Translations {
					if translation.Language == nil {
						continue
					}

					for _, configTranslation := range configEntry.Translations {
						if translation.Language.Name == configTranslation.Language {
							translationUpdate := make(map[string]interface{})

							if translation.SenderName != configTranslation.SenderName {
								translationUpdate["senderName"] = configTranslation.SenderName
							}

							if translation.Subject != configTranslation.Subject {
								translationUpdate["subject"] = configTranslation.Subject
							}

							if configTranslation.HTML != "" {
								if content, err := ioutil.ReadFile(configTranslation.HTML); err == nil {
									if translation.ContentHtml != string(content) {
										translationUpdate["contentHtml"] = string(content)
									}
								} else {
									log.Errorf("Cannot read file %s, with error: %s", configTranslation.HTML, err)
								}
							}

							if configTranslation.Plain != "" {
								if content, err := ioutil.ReadFile(configTranslation.Plain); err == nil {
									if translation.ContentPlain != string(content) {
										translationUpdate["contentPlain"] = string(content)
									}
								} else {
									log.Errorf("Cannot read file %s, with error: %s", configTranslation.Plain, err)
								}
							}

							localCustomFields, _ := json.Marshal(configTranslation.CustomFields)
							remoteCustomFields, _ := json.Marshal(translation.CustomFields)

							if string(localCustomFields) != string(remoteCustomFields) {
								translationUpdate["customFields"] = configTranslation.CustomFields
							}

							if len(translationUpdate) > 0 {
								translationUpdates[translation.LanguageId] = translationUpdate
							}
						}
					}

					if len(translationUpdates) > 0 {
						mailUpdate["translations"] = translationUpdates
					}
				}

				if len(mailUpdate) > 1 {
					mailUpdates = append(mailUpdates, mailUpdate)
				}
			}
		}
	}

	if len(mailUpdates) > 0 {
		operation.Operations["update-mail-template"] = adminSdk.SyncOperation{
			Action:  "upsert",
			Entity:  "mail_template",
			Payload: mailUpdates,
		}
	}

	return nil
}

func (s MailTemplateSync) Pull(ctx adminSdk.ApiContext, client *adminSdk.Client, config *shop.Config) error {
	mailTemplates, err := fetchAllMailTemplates(ctx, client)
	if err != nil {
		return err
	}

	config.Sync.MailTemplate = make([]shop.MailTemplate, 0)

	for _, row := range mailTemplates.Data {
		if row.MailTemplateType == nil {
			log.Infof("mail_template entity with id %s does not have a type. Skipping", row.Id)
			continue
		}

		cfg := shop.MailTemplate{
			Id:           row.Id,
			Translations: []shop.MailTemplateTranslation{},
		}

		for _, translation := range row.Translations {
			if translation.Language == nil {
				continue
			}

			configKey := translation.Language.Name

			htmLFilePath := fmt.Sprintf(".shopware-cli/mail-template/%s/%s-html.twig", row.MailTemplateType.TechnicalName, configKey)
			plainFilePath := fmt.Sprintf(".shopware-cli/mail-template/%s/%s-plain.twig", row.MailTemplateType.TechnicalName, configKey)
			dir := filepath.Dir(htmLFilePath)

			if _, err := os.Stat(dir); os.IsNotExist(err) {
				if createErr := os.MkdirAll(dir, os.ModePerm); createErr != nil {
					return createErr
				}
			}

			cfgLang := shop.MailTemplateTranslation{
				Language:     configKey,
				SenderName:   translation.SenderName,
				Subject:      translation.Subject,
				HTML:         htmLFilePath,
				Plain:        plainFilePath,
				CustomFields: translation.CustomFields,
			}

			if err := ioutil.WriteFile(htmLFilePath, []byte(translation.ContentHtml), os.ModePerm); err != nil {
				return err
			}

			if err := ioutil.WriteFile(plainFilePath, []byte(translation.ContentPlain), os.ModePerm); err != nil {
				return err
			}

			cfg.Translations = append(cfg.Translations, cfgLang)
		}

		config.Sync.MailTemplate = append(config.Sync.MailTemplate, cfg)
	}

	return nil
}

func fetchAllMailTemplates(ctx adminSdk.ApiContext, client *adminSdk.Client) (*adminSdk.MailTemplateCollection, error) {
	criteria := adminSdk.Criteria{}
	criteria.Includes = map[string][]string{
		"mail_template":             {"id", "mailTemplateType", "translations"},
		"mail_template_type":        {"technicalName"},
		"mail_template_translation": {"senderName", "subject", "contentHtml", "contentPlain", "language", "languageId"},
		"language":                  {"name"},
	}
	criteria.Associations = map[string]adminSdk.Criteria{"mailTemplateType": {}, "translations": {Associations: map[string]adminSdk.Criteria{"language": {}}}}

	collection, resp, err := client.Repository.MailTemplate.SearchAll(ctx, criteria)

	if err == nil {
		defer resp.Body.Close()
	}

	return collection, err
}
